package PPD;

use strict;
use warnings;

use Exporter 'import';
our @EXPORT = qw(get set maybe_set option generate);

my %options = ();

sub format_value($$) {
	my ($type, $value) = @_;
	$type //= 'q';
	if ($type eq 'q' || $type eq 'i') {
		# Quoted value / Invocation value
		if ($value =~ /["\x00-\x08\x0b-\x1f\x7f-\xff]/) { die "Invalid value: $value\n"; }
		$value = '"' . $value . '"';
		if ($value =~ /\n/) { $value .= "\n*End"; }
		return $value;
	} elsif ($type eq 's') {
		# String value
		return $value;
	} elsif ($type eq 'b') {
		# Boolean value
		return $value ? "True" : "False";
	} else {
		die "Unknown type '$type'\n";
	}
}

# get("section/option")
sub get($) {
	my ($key, $value) = @_;
	my ($section, $opt) = split /\//, $key;
	my $o = $options{$section}{$opt};
	if ($o) {
		return $o->{'value'};
	} else {
		return;
	}
}

# set("section/option[:type]", value)
# types: q=quoted (default), i=invocation, s=string (unquoted), b=boolean
sub set($$) {
	my ($key, $value) = @_;
	my ($section, $opt) = split /\//, $key;
	my ($opt2, $type) = split /:/, $opt;
	$options{$section}{$opt2} = { 'type' => $type, 'value' => $value };
}

sub maybe_set($$) {
	my ($key, $value) = @_;
	my ($section, $opt) = split /\//, $key;
	my ($opt2, $type) = split /:/, $opt;
	$options{$section}{$opt2} //= { 'type' => $type, 'value' => $value };
}

sub mand($) {
	my ($key) = @_;
	my ($section, $opt) = split /\//, $key;
	$options{$section}{$opt} = undef;
}

# Section "f": file description
set('f/FormatVersion', '4.3');
set('f/LanguageEncoding:s', 'ISOLatin1');
set('f/LanguageVersion:s', 'English');
mand('f/FileVersion');
mand('f/PCFileName');

# Section "p": product description
mand('p/Manufacturer');
mand('p/ModelName');
mand('p/NickName');
mand('p/ShortNickName');
mand('p/Product');
mand('p/PSVersion');

# Section "d": device capabilities
set('d/ColorDevice:b', 0);
set('d/DefaultColorSpace:q', 'Gray');
set('d/FileSystem:b', 0);

# Section "c": CUPS options
# (none by default)

# Section "j": JCL options
# (none by default)

my @ui_groups = ();

sub ui_group($) {
	my ($g) = @_;
	my $key = $g->{'Key'};
	defined $options{'u'}{$key} and die "Duplicate UI group $key\n";
	$options{'u'}{$key} = $g;
	push @ui_groups, $g;
	$g->{'Options'} = [];
}

# Section "u": UI groups
ui_group({ Key => '' });

# Section "n": non-UI options
$options{'n'} = { Key => '', Options => [] };

# option({
#	Key => key,		# (mandatory)
#	Name => name,		# descriptive name (default: same as Key)
#	Type => type,		# PickOne (default) / PickMany / Boolean
#	UI => group,		# UI group ('' if outside groups, default: non-UI option)
#	JCL => 1,		# if this is a JCL option
#	Section => sec,		# OrderDependency section: ExitServer / Prolog / DocumentSetup /
#				# PageSetup / JCLSetup (default if JCL=1) / AnySetup (default if JCL=0)
#	Priority => pri,	# OrderDependency priority, default = 100
#	Values => [		# possible values (mandatory)
#		{ Key => key,		# key (mandatory)
#		  Name => key,		# descriptive name (default: same as Key)
#		  PS => string,		# PS code to emit (mandatory)
#		  Default => 1,		# if this is the default value
#		},
#	]
# })
sub option($) {
	my ($o) = @_;
	my $ui = $o->{'UI'};
	my $g;
	if (defined $ui) {
		$g = $options{'u'}{$ui} or die "No such UI group $ui\n";
	} else {
		$g = $options{'n'};
	}
	push @{$g->{'Options'}}, $o;
}

sub heading($) {
	my ($h) = @_;
	print "\n";
	print "%* ", "=" x length $h, "\n";
	print "%* ", $h, "\n";
	print "%* ", "=" x length $h, "\n";
	print "\n";
}

sub emit($;$) {
	my ($section, $heading) = @_;
	my $opts = $options{$section};
	$opts or return;
	heading($heading) if defined $heading;
	for my $k (sort keys %$opts) {
		my $v = $opts->{$k};
		if (!$v) {
			print STDERR "ERROR: Value of mandatory option $k is missing\n";
		} else {
			print "*$k: ", format_value($v->{'type'}, $v->{'value'}), "\n";
		}
	}
}

sub get_default($) {
	my ($o) = @_;
	for my $val (@{$o->{Values}}) {
		return $val->{Key} if $val->{Default};
	}
	return $o->{Values}->[0]->{Key};
}

sub emit_option($) {
	my ($o) = @_;
	my $key = $o->{Key};
	my $type = $o->{Type} // 'PickOne';
	print "%* === $key ===\n";
	my $jcl = ($o->{JCL} // ($key =~ /^JCL/)) ? "JCL" : "";
	if (defined $o->{UI}) { print "*${jcl}OpenUI *$key/", ($o->{Name} // $o->{Key}), ": $type\n"; }
	my $pri = $o->{Priority} // 100;
	my $sec = $o->{Section} // ($jcl ? "JCLSetup" : "AnySetup");
	print '*', ((defined $o->{UI}) ? "" : 'NonUI'), "OrderDependency $pri $sec *$key\n";
	print "*Default$key: ", get_default($o), "\n";
	for my $v (@{$o->{Values}}) {
		my $ps = $v->{PS};
		# $ps = "\n$ps" unless $ps =~ /^\n/s;
		# $ps = "$ps\n" unless $ps =~ /\n$/s;
		print "*$key ", $v->{Key}, '/', ($v->{Name} // $v->{Key}), ': ', format_value('i', $ps), "\n";
	}
	if (defined $o->{UI}) { print "*${jcl}CloseUI *$key\n"; }
	print "\n";
}

sub emit_group($) {
	my ($g) = @_;
	if ($g->{Key} ne '') { print '*OpenGroup: ', $g->{Key}, '/', ($g->{Name} // $g->{Key}), "\n\n"; }
	for my $o (@{$g->{Options}}) {
		emit_option($o);
	}
	if ($g->{Key} ne '') { print "\n", '*CloseGroup: ', $g->{Key}, "\n"; }
}

sub fill_missing() {
	my $model = get('p/ModelName');
	if (defined $model) {
		maybe_set('p/Product', "($model)");
		maybe_set('p/NickName', $model);
		maybe_set('p/ShortNickName', $model);
	}

	my $psver = get('p/PSVersion');
	if (defined($psver) && $psver =~ /\((\d)/) {
		maybe_set('d/LanguageLevel', $1);
	}
}

sub generate() {
	print "*PPD-Adobe: ", get('f/FormatVersion'), "\n";
	print "*% PPD file generated by UCW PPD generator\n";
	fill_missing();

	emit('f', 'File version');
	emit('p', 'Product information');
	emit('d', 'Device capabilities');
	emit('c', 'CUPS options');
	emit('j', 'JCL setup');

	for my $g (@ui_groups) {
		next if $g->{Key} eq '';
		heading($g->{Name});
		emit_group($g);
	}

	heading("General UI options");
	emit_group($ui_groups[0]);

	if (@{$options{'n'}->{Options}}) {
		heading("Non-UI options");
		emit_group($options{'n'});
	}
}

42;
